#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAIN_C
#include "exitfail.h"

#define SERIAL_BAUDRATE B57600
#define BUF_SIZE 256

static int serial_fd = -1;
static struct termios old_tio;

static int terminated = 0;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) +1)
#define ARRAY_SIZE(a) (sizeof(a) /sizeof((a)[0]))

#define __unused __attribute__((unused))

/**** sample:Device ID:begin ****/
#define ID_BUF_SIZE 4

static unsigned char id_lock[ID_BUF_SIZE] = {0x00, 0x2b, 0xcc, 0x89};
static unsigned char id_open[ID_BUF_SIZE] = {0x04, 0x00, 0x50, 0x6c};
static unsigned char id_th[ID_BUF_SIZE] = {0x04, 0x00, 0x0c, 0x78};
static unsigned char id_jin[ID_BUF_SIZE] = {0x04, 0x00, 0x18, 0x39};

enum SENS_KIND {
	SENS_NONE = 0,
	SENS_LOCK = 1,
	SENS_OPEN = 2,
	SENS_TH = 3,
	SENS_JIN = 4,
};
/**** sample:Device ID:end ****/

static void terminate_sig_handler(__unused int signum)
{
	terminated = 1;
}

static void set_sig_handler(int sig_list[], ssize_t num, __sighandler_t handler)
{
	struct sigaction sa;
	int i;
	
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;

	for(i=0; i<num; i++)
		if(sigaction(sig_list[i], &sa, NULL) < 0)
			exitfail_errno("sigaction");
}

static void restore_serial(void)
{
	int ret;

	ret = tcsetattr(serial_fd, TCSANOW, &old_tio);
	if(ret < 0) exitfail_errno("tcsetattr");
}

static void setup_serial(int fd)
{
	struct termios tio;
	int ret;

	ret = tcgetattr(fd, &old_tio);
	if(ret) exitfail_errno("tcsgetattr");

	if(atexit(restore_serial)) exitfail_errno("atexit");

	memset(&tio, 0, sizeof(tio));
	tio.c_iflag = IGNBRK | IGNPAR;
	tio.c_cflag = CS8 | CLOCAL | CREAD;
	tio.c_cc[VTIME] = 0;
	tio.c_cc[VMIN] = 1;
	ret = cfsetspeed(&tio, SERIAL_BAUDRATE);
	if(ret < 0) exitfail_errno("cfsetspeed");

	ret = tcflush(fd, TCIFLUSH);
	if(ret < 0) exitfail_errno("tcflush");

	ret = tcsetattr(fd, TCSANOW, &tio);
	if(ret) exitfail_errno("tcsetattr");
}

int main(int argc, char *argv[])
{
	int terminate_sig_list[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};
//	char buf[BUF_SIZE]; // original:co
	ssize_t ret, len, wrlen;
/**** sample:variable:begin ****/
       unsigned char buf[BUF_SIZE];
	unsigned char buf1[BUF_SIZE];
	unsigned char id_buf[ID_BUF_SIZE];
       int cnt_char;
       int cnt_head;
	int cnt_esp;
	int data_len;
	int ti_flag;
	int i;
	float data_tmp;
	float data_hum;
	ssize_t analen;
	enum SENS_KIND sens_kind1;
	FILE *fp;
	int cnt_sw84, cnt_sw88;
	int cnt_open;
	int cnt_jin;
/**** sample:variable:end ****/

	exitfail_init();

	if(argc < 2) {
		printf("Usage: %s <device>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	serial_fd = open(argv[1], O_RDWR | O_NOCTTY);
	if(serial_fd < 0) exitfail_errno("open");

	set_sig_handler(terminate_sig_list, ARRAY_SIZE(terminate_sig_list),
			terminate_sig_handler);

	setup_serial(serial_fd);

/**** sample:initial value:begin ****/
	cnt_char = 0;
	cnt_head = 0;
	cnt_esp = 0;
	data_len = 0;
	ti_flag = 0;
	data_tmp = 0;
	data_hum = 0;
	sens_kind1 = SENS_NONE;
	for(i = 0; i < ID_BUF_SIZE; i++) id_buf[i] = 0x00;
	cnt_sw84 = 0;
	cnt_sw88 = 0;
	cnt_open = 0;
	cnt_jin = 0;
/**** sample:initial value:end ****/

	while(!terminated) {
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) {
			if(errno == EINTR) continue;
			exitfail_errno("read");
		}
		len = ret;

/**** sample:receive:begin ****/
		for(analen = 0; analen < len; analen++) {
			if(cnt_head == 0 && buf[analen] == 0x55) {
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head == 1 && buf[analen] == 0x00) {
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head == 2 && (buf[analen] == 0x07 
							|| buf[analen] == 0x09
							|| buf[analen] == 0x0a
							|| buf[analen] == 0x0c)) {
				if(buf[analen]==0x0c) ti_flag = 1;
				else ti_flag = 0;
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				cnt_head++;
				data_len = buf1[1] * 256 + buf1[2];
			}
			else if(cnt_head == 3 && buf[analen] == 0x02) {
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head == 4 && buf[analen] == 0x0a) {
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head == 5) { /**** not CRC check ****/
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head == 6) { /**** not Data header check ****/
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head > 6 && cnt_esp < 6 + data_len + 3) {
				buf1[cnt_esp] = buf[analen];
				cnt_esp++;
				if(cnt_esp == (6 + data_len + 3)) {
					if(buf1[6] == 0x61 || buf1[6] == 0x62) ti_flag = 1;
					else ti_flag = 0;

					if(ti_flag == 0) 
						for(i = 0; i < ID_BUF_SIZE; i++) 
							id_buf[i] = buf1[7 + i]; 
					else 
						for(i = 0; i < ID_BUF_SIZE; i++) 
							id_buf[i] = buf1[9 + i]; 

					if (memcmp(id_buf, id_lock, 4) == 0) {
						sens_kind1 = SENS_LOCK;
						fp = fopen("/root/log_lock.txt", "w");
					}
					else if (memcmp(id_buf, id_open, 4) == 0) {
						sens_kind1 = SENS_OPEN;
						fp = fopen("/root/log_open.txt", "w");
					}
					else if (memcmp(id_buf, id_th, 4) == 0) {
						sens_kind1 = SENS_TH;
						fp = fopen("/root/log_th.txt", "w");
					}
					else if (memcmp(id_buf, id_jin, 4) == 0) {
						sens_kind1 = SENS_JIN;
						fp = fopen("/root/log_jin.txt", "w");
					}
					else sens_kind1 = SENS_NONE;

					if(ti_flag == 0 && sens_kind1 != SENS_NONE) {
						fprintf(fp, "ID:");
						for(i = 0; i < ID_BUF_SIZE; i++) {
							fprintf(fp, "%02x", id_buf[i]);
						}
						fprintf(fp, ": ");

						/**** calculate temparature/humidity ****/
						if(sens_kind1 == SENS_LOCK) {
							fprintf(fp, "LockerSW: ");
							if(buf1[11] == 0x84) cnt_sw84++;
							else if(buf1[11] == 0x88) cnt_sw88++;
							fprintf(fp, "Switch84:%d, Switch88:%d\n"
								, cnt_sw84, cnt_sw88);
						}
						else if(sens_kind1 == SENS_OPEN) {
							fprintf(fp, "MagnetSW: ");
							if(buf1[11] == 0x08) {
								cnt_open++;
								fprintf(fp, "open:%d, status:open\n", cnt_open);
							}
							else if(buf1[11] == 0x09) 
								fprintf(fp, "open:%d, status:close\n", cnt_open);
						}
						else if(sens_kind1 == SENS_TH) {
							fprintf(fp, "TEMP-HUM: ");
							data_hum = (float)buf1[11] * 100 / 255;
							data_tmp = ((float)(buf1[12] & 0x03) * 256 
							+ (float)buf1[13]) * (60 - (-20)) / 1023 
							+ (-20);
							fprintf(fp, "%.2f[degree], %.2f[percent]\n"
								, data_tmp, data_hum);
						}
						else if(sens_kind1 == SENS_JIN) {
							fprintf(fp, "JINKAN: ");
							if(buf1[13] == 0xFF) {
								cnt_jin++;
								fprintf(fp, "detect:%d, status:detect\n", cnt_jin);
							}
							else if(buf1[13] == 0x00) 
								fprintf(fp, "detect:%d, status:not detect\n", cnt_jin);
						}

						fclose(fp);
					}
					else if(ti_flag == 1 && sens_kind1 != SENS_NONE) {
						printf("TEACH-IN(ID):");
						for(i = 7; i < 7 + 6; i++) {
							printf("%02x", buf1[i]);
						}
						printf(": ");

						printf("TEACH-IN(Data):");
						for(i = 13; i < 13 + 4; i++) {
							printf("%02x", buf1[i]);
						}
						printf("\n");

						fclose(fp); /**** not output to log_*.txt ****/
					}

					cnt_head = 0;
					cnt_esp = 0;
					data_len = 0;
				}
			}
			else {
				cnt_head = 0;
				cnt_esp = 0;
				data_len = 0;
			}
		}
/**** sample:receive:end ****/

/**** sample:add:begin for debug ****/
/*
		for(wrlen = 0; wrlen < len; wrlen++) {
			printf("%02x", buf[wrlen]);
			cnt_char++;
			if(cnt_char==16) {
				printf("\n");
				cnt_char = 0;
			}
		}
*/
/**** sample:add:end for debug ****/
	}
	return EXIT_SUCCESS;
}

